import fs from "node:fs";
import { inspect as _inspect } from "node:util";
import * as emphasize from "emphasize";
import * as termKit from "terminal-kit";
import * as zx from "zx";
const args = process.argv.slice(2);
const inspectOptions = {
  colors: true,
  depth: Infinity
};
const inspect = (value) => _inspect(value, inspectOptions);
const debugStream = process.env.NODE_ENV === "development" ? fs.createWriteStream("debug.txt") : {
  write() {
  },
  close() {
  }
};
const run = async (code) => {
  const result = await eval(`(async function() { ${code} }).bind(global)()`);
  return result;
};
function runCode(code2) {
  return new Promise((resolve, reject) => {
    const timer = setTimeout(() => reject(new Error("Timeout")), 3e3);
    (async () => {
      const isExpression = !code2.includes(";");
      try {
        const result2 = await run(isExpression ? `return ${code2}` : code2);
        clearTimeout(timer);
        resolve(result2);
      } catch (e) {
        if (!isExpression) {
          clearTimeout(timer);
          reject(e);
          return;
        }
        run(code2).then(resolve).catch(reject).finally(() => clearTimeout(timer));
      }
    })().catch(console.error);
  });
}
const userAgent = (() => {
  if (typeof window === "undefined")
    return void 0;
  const ua = window?.navigator?.userAgent ?? "";
  if (ua !== "") {
    return ua.includes("Deno") ? void 0 : ua;
  }
  return void 0;
})();
const getProcess = () => {
  try {
    return !userAgent && typeof process !== "undefined" ? process : {};
  } catch (err) {
    return {};
  }
};
const bunVersion = getProcess()?.versions?.bun;
const isBun = Boolean(bunVersion);
const denoVersion = (
  // eslint-disable-next-line no-undef
  typeof Deno !== "undefined" ? Deno?.version?.deno : void 0
);
const isDeno = Boolean(denoVersion) && !isBrowser;
const nodeVersion = isBun || isDeno ? void 0 : getProcess()?.versions?.node;
const isNode = Boolean(nodeVersion) && !isBun && !isDeno;
async function prepareRuntime() {
  Object.assign(globalThis, zx, {
    /**
     * Manually pipe key inputs because zx.$ does something to stdin that stops keypress
     * events after it exists.
     */
    $: async function(pieces, ...args2) {
      const p = zx.$(pieces, ...args2).nothrow().timeout("5s").stdio("pipe", "pipe", "pipe").quiet();
      const result2 = await p;
      if (result2.exitCode) {
        return new Error(result2.stderr || `Exit code ${result2.exitCode}`);
      }
      return result2.toString().trim();
    },
    $see: async function(pieces, ...args2) {
      const p = zx.$(pieces, ...args2).nothrow().timeout("5s").stdio("pipe", "pipe", "pipe");
      const result2 = await p;
      if (result2.exitCode) {
        return new Error(result2.stderr || `Exit code ${result2.exitCode}`);
      }
      return result2.toString().trim();
    },
    inspect,
    log: console.log,
    globDir: (pattern, options = {}) => zx.glob(pattern, {
      onlyDirectories: true,
      ...options
    }),
    exit: () => process.exit(),
    ...isBun ? Object.assign({}, await import("bun:sqlite")) : {}
  });
}
const emph = emphasize.createEmphasize(emphasize.common);
debugStream.write("\n");
process.on("beforeExit", () => {
  debugStream.close();
});
let term = globalThis.term;
if (term) {
  term.removeAllListeners();
} else {
  term = globalThis.term = termKit.terminal;
}
let buffer = "";
let startPosition = { x: 0, y: 0 };
let endPosition = { x: 0, y: 0 };
let bufferCursorIndex = 0;
let prompt = "";
let promptLength = 0;
const history = [];
let historyCursor = 0;
function setPrompt(str) {
  prompt = str;
  promptLength = termKit.stringWidth(str);
}
async function showPrompt() {
  Object.assign(startPosition, await term.getCursorLocation());
  debugStream.write("Prompt start: " + JSON.stringify(startPosition) + "\n");
  term(prompt);
  Object.assign(endPosition, {
    x: startPosition.x + promptLength,
    y: startPosition.y
  });
  bufferCursorIndex = 0;
}
function calculateCursorPosition(index = bufferCursorIndex, relative = false) {
  index += promptLength;
  const row = Math.floor(index / term.width);
  let col = index - term.width * row;
  if (col < 0)
    col = 0;
  let x = col;
  let y = row;
  if (!relative) {
    x += startPosition.x;
    y += startPosition.y;
  }
  const cursor = {
    x,
    y
  };
  return cursor;
}
async function updateLine() {
  const output = buffer ? emph.highlight("js", buffer).value : "";
  term.eraseArea(
    startPosition.x,
    startPosition.y,
    term.width,
    endPosition.y - startPosition.y + 1
  );
  term.hideCursor(true);
  term.moveTo(startPosition.x, startPosition.y);
  term(`\r${prompt}${output}`);
  const relativeEndPosition = calculateCursorPosition(buffer.length, true);
  const endRow = startPosition.y + relativeEndPosition.y;
  if (endRow >= term.height) {
    const diff = term.height - endRow;
    debugStream.write(`Scroll: ${diff}
`);
    startPosition.y += diff;
    if (diff === -1 && relativeEndPosition.x === 0) {
      term("\n");
    }
  }
  Object.assign(endPosition, await term.getCursorLocation());
  debugStream.write(
    "DRAW " + JSON.stringify(startPosition) + " ~ " + JSON.stringify(endPosition) + "\n"
  );
  const cursor = calculateCursorPosition();
  term.moveTo(cursor.x, cursor.y);
  term.hideCursor(false);
  debugStream.write(
    "CURSOR " + JSON.stringify(calculateCursorPosition()) + " = " + JSON.stringify(bufferCursorIndex) + "\n"
  );
}
function clearScreen() {
  term.clear();
  Object.assign(startPosition, { x: 1, y: 1 });
  updateLine();
}
term.on("key", async (name, matches, data) => {
  if (data.isCharacter) {
    term(name);
    buffer = buffer.slice(0, bufferCursorIndex) + name + buffer.slice(bufferCursorIndex);
    bufferCursorIndex++;
    updateLine();
    return;
  }
  debugStream.write(JSON.stringify([name, data]) + "\n");
  switch (name) {
    case "CTRL_C":
      term("\n");
      process.exit();
      break;
    case "BACKSPACE":
      if (buffer[bufferCursorIndex - 1]) {
        buffer = buffer.slice(0, bufferCursorIndex - 1) + buffer.slice(bufferCursorIndex);
        bufferCursorIndex--;
        updateLine();
      }
      break;
    case "DELETE":
      if (buffer[bufferCursorIndex]) {
        buffer = buffer.slice(0, bufferCursorIndex) + buffer.slice(bufferCursorIndex + 1);
        updateLine();
      }
      break;
    case "ENTER":
      {
        const cursor = calculateCursorPosition(buffer.length);
        term.moveTo(cursor.x, cursor.y);
      }
      term("\n");
      if (buffer === "help") {
        console.log(`Available commands`);
        console.log(`clear - Clear screen`);
        console.log(`exit - Exit`);
        console.log(`help - Show this help screen`);
        buffer = "";
        showPrompt();
        return;
      } else if (buffer === "clear") {
        buffer = "";
        bufferCursorIndex = 0;
        clearScreen();
        return;
      } else if (buffer === "exit") {
        process.exit();
        return;
      }
      history.push(buffer);
      history.splice(100, history.length - 100);
      historyCursor = history.length;
      debugStream.write(`EVAL: ${buffer}
`);
      try {
        const result2 = await runCode(buffer);
        if (result2) {
          console.log(typeof result2 === "object" ? inspect(result2) : result2);
        }
      } catch (e) {
        term.red(e.message + "\n");
      }
      buffer = "";
      showPrompt();
      break;
    case "UP":
      if (historyCursor > 0 && history[historyCursor - 1]) {
        buffer = history[historyCursor - 1];
        historyCursor--;
        bufferCursorIndex = buffer.length;
        updateLine();
      }
      break;
    case "DOWN":
      debugStream.write(
        `DOWN: ${historyCursor + 1} ? ${history[historyCursor + 1]}
`
      );
      if (history.length === 0)
        return;
      if (history[historyCursor + 1]) {
        buffer = history[historyCursor + 1];
        historyCursor++;
        bufferCursorIndex = buffer.length;
      } else {
        historyCursor = history.length;
        buffer = "";
        bufferCursorIndex = 0;
      }
      updateLine();
      break;
    case "LEFT":
      if (bufferCursorIndex > 0) {
        bufferCursorIndex--;
        const cursor = calculateCursorPosition();
        term.moveTo(cursor.x, cursor.y);
        debugStream.write(`LEFT: ${JSON.stringify(cursor)}
`);
      }
      break;
    case "RIGHT":
      if (bufferCursorIndex < buffer.length) {
        bufferCursorIndex++;
        const cursor = calculateCursorPosition();
        term.moveTo(cursor.x, cursor.y);
        debugStream.write(`RIGHT: ${JSON.stringify(cursor)}
`);
      }
      break;
    case "CTRL_LEFT":
    case "CTRL_A":
      {
        bufferCursorIndex = 0;
        const cursor = calculateCursorPosition();
        term.moveTo(cursor.x, cursor.y);
      }
      break;
    case "CTRL_RIGHT":
    case "CTRL_E":
      {
        bufferCursorIndex = buffer.length;
        const cursor = calculateCursorPosition();
        term.moveTo(cursor.x, cursor.y);
      }
      break;
    case "CTRL_R":
      clearScreen();
      break;
    default:
      break;
  }
  if (name === "CTRL_C")
    process.exit();
});
setPrompt("\x1B[033\x1B[01;34mzx\x1B[033\x1B[01;30m>\x1B[033\x1B[00m ");
console.log("\x1B[1;32mzxel\x1B[00m - Interactive JavaScript runtime shell");
console.log(
  '\x1B[2;37mEnter code to run, or "help" to see commands\nAssign variables like: this.a = 123\x1B[00m'
);
prepareRuntime().then(() => {
  term.grabInput(true);
  showPrompt();
});
export {
  bunVersion,
  denoVersion,
  isBun,
  isDeno,
  isNode,
  nodeVersion
};
